// Last edited on 2011-04-23 19:55:53 by stolfi
// Fukushima Daiichi - generic macros (for any unit)
    
#macro generic_flanged_capped_cylinder(H,R,D,T,FR,FT,bc,tc) 
  // A vertical cylindrical vessel terminated by hemiellipsoidal caps or flanges.
  // The cylindical part has length {H} and outer radius {R}.
  // Each dome has height {D}.
  // The flanges have vertical thickness {FT} and extend by {FR} outwards.
  // The walls have thickness {T}.
  // The bottom is dome if {bc=1}, flange if {bc=0}.
  // The top    is dome if {tc=1}, flange if {tc=0}.
  // The origin is on the axis at the height of the bottom seam.
                                                   
  #local bot_ctr = < 0, 0, 0 >;
  #local top_ctr = < 0, 0, H >;
  union{
    #if (bc > 0)
      object{ generic_vessel_dome(R,D,T) scale <1,1,-1> }
    #else
      object{ generic_vessel_flange(R,T,FR,FT) }
    #end
    difference{
      cylinder{ 
        #if (bc > 0) bot_ctr + eps*z #else bot_ctr - eps*z #end, 
        #if (tc > 0) top_ctr + eps*z #else top_ctr - eps*z #end,
        R - eps
      }
      cylinder{ bot_ctr - 3*eps*z, top_ctr + 3*eps*z, R - T + eps }
    }
    #if (tc > 0)
      object{ generic_vessel_dome(R,D,T) translate top_ctr }
    #else
      object{ generic_vessel_flange(R,T,FR,FT) scale <1,1,-1> translate top_ctr }
    #end
    
    bounded_by{ 
      cylinder{ 
        #if (bc > 0) bot_ctr - D*z #else bot_ctr #end,
        #if (tc > 0) top_ctr + D*z #else top_ctr #end,
        #if ((bc > 0) | (tc > 0)) R + FR #else R #end
      }
    }
  }
#end

#macro generic_vessel_dome(R,D,T)
  // Hemiellipsoidal dome with height {D} for closing a cylindrical vessel of
  // of outer radius {R} and thickness {T}.
  // Sitting on {Z=0} plane, dome up, centered on {Z} axis.
  #local ot_scZ = (D-eps)/(R-eps);     // Z flattening factor for outer surf of dome.
  #local in_scZ = (D-T+eps)/(R-T+eps); // Z flattening factor for inner surf of dome.
  #local ctr = <0,0,0>;
  intersection{
    cylinder{ ctr - eps*z, ctr + D*z, R }
    difference{
      sphere{ ctr, R-eps scale <1,1,ot_scZ> }
      sphere{ ctr, R-T+eps scale <1,1,in_scZ> }
    }
    bounded_by{ cylinder{ ctr - eps*z, ctr + D*z, R } }
  }
#end

#macro generic_vessel_flange(R,T,FR,FT)
  // Flange for a round tube of outer radius {R} and thickness {T}.
  // The flange has thckness {FT} and extends out by {FR}.
  // Sits on {Z=0} plane centered on {Z} axis.
  #local bot_ctr = <0,0,0>;
  #local top_ctr = bot_ctr + FT*z;
  difference{
    cylinder{ bot_ctr + eps*z, top_ctr - eps*z, R + FR - eps }
    cylinder{ bot_ctr - eps*z, top_ctr + eps*z, R - T + 2*eps }
    bounded_by{ cylinder { bot_ctr, top_ctr, R+FR } }
  }
#end
 
#macro generic_pool_cavity( lo_X,sz_X,lo_Y,sz_Y,lo_Z,hi_Z, ch_wid,ch_len,ch_dir,ch_lo_Z)
  // An negative object to excavate a pool on some floor, and its connecting channel.
  // The channel extends from the center of the pool's edge in direction {ch_dir}.
  
  // {lo_Y} = Min X, rel west wall.
  // {sz_X} = X size.
  // {lo_Y} = Min Y, rel south wall.
  // {sz_Y} = Y size.
  // {lo_Z} = Z of bottom of pool, rel OP.
  // {hi_Z} = Z of top of pool (= room floor), rel OP.
  
  // {ch_wid} = Width of channel.
  // {ch_len} = Length of channel.
  // {ch_lo_Z} = Z of bottom of channel, rel OP.
  
  // Coordinates of the pool cavity proper:
  #local hi_X = lo_X + sz_X;             // Max X, rel east wall..
  #local hi_Y = lo_Y + sz_Y;             // Max Y, rel south wall.

  #local lo_corner = < lo_X, lo_Y, lo_Z >;
  #local hi_corner = < hi_X, hi_Y, hi_Z + 0.500 >;

  // Coordinates of the channel between pool and parent cavity:
  #if (ch_dir.y = 0)
    #local ch_md_Y = (lo_Y + hi_Y)/2;         // Center Y (assuming symmetrical) [TBC].
    #local ch_lo_Y = ch_md_Y - ch_wid/2;      // Min Y, rel south wall.
    #local ch_hi_Y = ch_lo_Y + ch_wid;        // Max Y, rel south wall.
    #if (ch_dir.x > 0)                        //
      #local ch_lo_X = hi_X;                  // Min X, rel east wall.
      #local ch_hi_X = ch_lo_X + ch_len;      // Max X, rel east wall.
    #else
      #local ch_hi_X = lo_X;                  // Max X, rel east wall.
      #local ch_lo_X = ch_hi_X - ch_len;      // Min X, rel east wall.
    #end
  #end
  #if (ch_dir.x = 0)
    #local ch_md_X = (lo_X + hi_X)/2;        // Center X (assuming symmetrical) [TBC].
    #local ch_lo_X = ch_md_X - ch_wid/2;  // Min X, rel east wall..
    #local ch_hi_X = ch_lo_X + ch_wid;    // Max X, rel east wall..
    #if (ch_dir.y > 0)                        //
      #local ch_lo_Y = hi_Y;                  // Min Y, rel east wall.
      #local ch_hi_Y = ch_lo_Y + ch_len;    // Max Y, rel east wall.
    #else
      #local ch_hi_Y = lo_Y;                  // Max Y, rel east wall.
      #local ch_lo_Y = ch_hi_Y - ch_len;    // Min Y, rel east wall.
    #end
  #end
    
  #local ch_hi_Z = hi_Z;      // Max Z, rel OP.

  #local ch_lo_corner = < ch_lo_X, ch_lo_Y, ch_lo_Z >;
  #local ch_hi_corner = < ch_hi_X, ch_hi_Y, ch_hi_Z + 0.500 >;
  
  union{
    box{ lo_corner - epsv, hi_corner + epsv }
    box{ ch_lo_corner - epsv, ch_hi_corner + epsv }
  }
#end

#macro generic_pool_cavity_walls( lo_X,sz_X,lo_Y,sz_Y,lo_Z,hi_Z,thk, ch_wid,ch_len,ch_dir,ch_lo_Z,ch_thk)
  // A positive object to create a block from which one may excavate a {generic_pool_cavity}
  // to obtain the walls of the pool and of its connecting channel.
  // The channel extends from the center of the pool's edge in direction {ch_dir}.
  
  // {lo_Y} = Min X, rel west wall.
  // {sz_X} = X size.
  // {lo_Y} = Min Y, rel south wall.
  // {sz_Y} = Y size.
  // {lo_Z} = Z of bottom of pool, rel OP.
  // {hi_Z} = Z of top of pool (= room floor), rel OP.
  // {thk} = pool wall thickness.
  
  // {ch_wid} = Width of channel.
  // {ch_len} = Length of channel.
  // {ch_lo_Z} = Z of bottom of channel, rel OP.
  // {ch_thk} = channel wall thickness.
  
  #local T = thk;
  #local S = ch_thk;

  // Coordinates of the pool cavity proper:
  #local hi_X = lo_X + sz_X;             // Max X, rel east wall..
  #local hi_Y = lo_Y + sz_Y;             // Max Y, rel south wall.

  #local lo_corner = < lo_X, lo_Y, lo_Z >;
  #local hi_corner = < hi_X, hi_Y, hi_Z >;

  // Coordinates of the channel between pool and parent cavity:
  #if (ch_dir.y = 0)
    #local ch_md_Y = (lo_Y + hi_Y)/2;         // Center Y (assuming symmetrical) [TBC].
    #local ch_lo_Y = ch_md_Y - ch_wid/2;      // Min Y, rel south wall.
    #local ch_hi_Y = ch_lo_Y + ch_wid;        // Max Y, rel south wall.
    #if (ch_dir.x > 0)                        //
      #local ch_lo_X = hi_X;                  // Min X, rel east wall.
      #local ch_hi_X = ch_lo_X + ch_len;      // Max X, rel east wall.
    #else
      #local ch_hi_X = lo_X;                  // Max X, rel east wall.
      #local ch_lo_X = ch_hi_X - ch_len;      // Min X, rel east wall.
    #end
  #end
  #if (ch_dir.x = 0)
    #local ch_md_X = (lo_X + hi_X)/2;         // Center X (assuming symmetrical) [TBC].
    #local ch_lo_X = ch_md_X - ch_wid/2;      // Min X, rel east wall..
    #local ch_hi_X = ch_lo_X + ch_wid;        // Max X, rel east wall..
    #if (ch_dir.y > 0)                        //
      #local ch_lo_Y = hi_Y;                  // Min Y, rel east wall.
      #local ch_hi_Y = ch_lo_Y + ch_len;      // Max Y, rel east wall.
    #else
      #local ch_hi_Y = lo_Y;                  // Max Y, rel east wall.
      #local ch_lo_Y = ch_hi_Y - ch_len;      // Min Y, rel east wall.
    #end
  #end
    
  #local ch_hi_Z = hi_Z;      // Max Z, rel OP.

  #local ch_lo_corner = < ch_lo_X, ch_lo_Y, ch_lo_Z >;
  #local ch_hi_corner = < ch_hi_X, ch_hi_Y, ch_hi_Z >;
  
  union{
    #if (T > 0)
      box{ lo_corner - <T,T,T> + epsv, hi_corner + <T,T,0> - epsv }
    #end
    #if (S > 0)
      box{ ch_lo_corner - <S,S,S> + epsv, ch_hi_corner + <S,S,0> - epsv }
    #end
  }
#end

#macro debug_coord(lab,vv)
  #debug "\n"
  #debug "!! "
  #debug lab
  #debug " = "
  #debug str(vv,8,5)
  #debug "\n"
#end

#macro debug_int(lab,vv)
  #debug "\n"
  #debug "!! "
  #debug lab
  #debug " = "
  #debug str(vv,1,0)
  #debug "\n"
#end

#macro debug_point(lab,pt)
  #debug "\n"
  #debug "!! "
  #debug lab
  #debug " = < "
  #debug str(pt.x,6,3)
  #debug " "
  #debug str(pt.y,6,3)
  #debug " "
  #debug str(pt.z,6,3)
  #debug " >\n"
#end
